import importlib

from django.apps import AppConfig, apps
from django.db.models.signals import post_delete


class AlreadyRegistered(Exception):
    pass


class BaseLinkcheckConfig(AppConfig):
    name = 'linkcheck'
    verbose_name = "Linkcheck"

    default_auto_field = "django.db.models.BigAutoField"

    all_linklists = {}

    def ready(self):
        self.build_linklists()

    def build_linklists(self):
        """Autodiscovery of linkLists"""
        for app in apps.get_app_configs():
            module_name = f"{app.name}.linklists"
            try:
                if not importlib.util.find_spec(module_name):
                    continue
            except ModuleNotFoundError:
                continue
            the_module = importlib.import_module(module_name)
            try:
                for k in the_module.linklists.keys():
                    if k in self.all_linklists.keys():
                        raise AlreadyRegistered(f'The key {k} is already registered in all_linklists')

                for link_list in the_module.linklists.values():
                    for link_list2 in self.all_linklists.values():
                        if link_list.model == link_list2.model:
                            raise AlreadyRegistered(f"The LinkList {link_list} is already registered in all_linklists")
                self.all_linklists.update(the_module.linklists)
            except AttributeError:
                pass
        # Add a reference to the linklist in the model. This change is for internal hash link,
        # But might also be useful elsewhere in the future
        for key, linklist in self.all_linklists.items():
            setattr(linklist.model, '_linklist', linklist)


class LinkcheckConfig(BaseLinkcheckConfig):
    default = True

    def ready(self):
        from .linkcheck_settings import DISABLE_LISTENERS
        from .listeners import register_listeners
        super().ready()

        if not DISABLE_LISTENERS:
            register_listeners()

        from .models import Link, link_post_delete
        post_delete.connect(link_post_delete, sender=Link)



from django.apps import apps
from django.core.management.base import BaseCommand, CommandError
from django.utils.termcolors import make_style

from linkcheck.utils import get_coverage_data, get_suggested_linklist_config


class Command(BaseCommand):

    cyan = staticmethod(make_style(fg='cyan'))

    help = 'Go through all models and check whether they are registered with linkcheck'

    def add_arguments(self, parser):
        parser.add_argument(
            '--model',
            help="Generate the suggested config for this model",
        )

    def handle(self, *args, model, **options):
        if model:
            try:
                model_class = apps.get_model(model)
            except Exception as e:
                raise CommandError(
                    f'Model "{model}" does not exist.'
                ) from e
            self.stdout.write(get_suggested_linklist_config(model_class))
        else:
            covered, uncovered = get_coverage_data()
            self.stdout.write('All covered models:\n')
            self.stdout.write(', '.join(map(self.cyan, covered)))
            for model, suggested_config in uncovered:
                self.stdout.write(f'\nSuggested config for model {model}:')
                self.stdout.write(self.cyan(suggested_config))



from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('linkcheck', '0003_redirect_to_as_textfield'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='url',
            name='still_exists',
        ),
    ]



import os
import sys
from datetime import datetime, timedelta
from io import StringIO
from unittest.mock import patch

import requests_mock
import urllib3
from django.apps import apps
from django.conf import settings
from django.contrib.auth.models import User
from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import LiveServerTestCase, TestCase
from django.test.utils import override_settings
from django.urls import reverse
from requests.exceptions import ConnectionError

from linkcheck.linkcheck_settings import MAX_URL_LENGTH
from linkcheck.listeners import (
    disable_listeners,
    enable_listeners,
    linkcheck_worker,
    register_listeners,
    tasks_queue,
    unregister_listeners,
)
from linkcheck.models import Link, Url
from linkcheck.views import get_jquery_min_js

from .sampleapp.models import Author, Book, Journal, Page


@override_settings(ROOT_URLCONF='linkcheck.tests.urls')
class InternalCheckTestCase(TestCase):

    def test_internal_check_mailto(self):
        uv = Url(url="mailto:nobody")
        uv.check_url()
        self.assertEqual(uv.message, 'Email link (not automatically checked)')
        self.assertEqual(uv.get_message, 'Email link (not automatically checked)')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, None)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, '')
        self.assertEqual(uv.get_status_code_display(), None)
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'mailto')

    def test_internal_check_tel(self):
        uv = Url(url="tel:+123456789")
        uv.check_url()
        self.assertEqual(uv.message, 'Phone number (not automatically checked)')
        self.assertEqual(uv.get_message, 'Phone number link (not automatically checked)')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, None)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, '')
        self.assertEqual(uv.get_status_code_display(), None)
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'phone')

    def test_internal_check_blank(self):
        uv = Url(url="")
        uv.check_url()
        self.assertEqual(uv.message, 'Empty link')
        self.assertEqual(uv.get_message, 'Empty link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, False)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, '')
        self.assertEqual(uv.get_status_code_display(), None)
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'empty')

    def test_same_page_anchor(self):
        uv = Url(url="#some_anchor")
        uv.check_url()
        self.assertEqual(uv.message, 'Link to within the same page (not automatically checked)')
        self.assertEqual(uv.get_message, 'Anchor link (not automatically checked)')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, None)
        self.assertEqual(uv.anchor_message, 'Anchor could not be checked')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, '')
        self.assertEqual(uv.get_status_code_display(), None)
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'anchor')

    def test_working_internal_anchor(self):
        uv = Url(url="/http/anchor/#anchor")
        uv.check_url()
        self.assertEqual(uv.message, "Working internal link, working internal hash anchor")
        self.assertEqual(uv.get_message, 'Working internal link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, 'Working anchor')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, '')
        self.assertEqual(uv.get_status_code_display(), '200 OK')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'internal')

    @patch("linkcheck.models.TOLERATE_BROKEN_ANCHOR", False)
    def test_broken_internal_anchor(self):
        uv = Url(url="/http/anchor/#broken-anchor")
        uv.check_url()
        self.assertEqual(uv.message, "Working internal link, broken internal hash anchor")
        self.assertEqual(uv.get_message, 'Working internal link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, False)
        self.assertEqual(uv.anchor_message, 'Broken anchor')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, '')
        self.assertEqual(uv.get_status_code_display(), '200 OK')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'internal')

    def test_broken_internal_anchor_tolerated(self):
        uv = Url(url="/http/anchor/#broken-anchor")
        uv.check_url()
        self.assertEqual(uv.message, "Working internal link, broken internal hash anchor")
        self.assertEqual(uv.get_message, 'Working internal link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, 'Broken anchor')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, '')
        self.assertEqual(uv.get_status_code_display(), '200 OK')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'internal')

    def test_redirect_working_internal_anchor(self):
        uv = Url(url="/http/redirect_to_anchor/#anchor")
        uv.check_url()
        self.assertEqual(uv.message, "Working temporary redirect, working internal hash anchor")
        self.assertEqual(uv.get_message, 'Working temporary redirect')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, 'Working anchor')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, '')
        self.assertEqual(uv.get_status_code_display(), '302 Found')
        self.assertEqual(uv.get_redirect_status_code_display(), '200 OK')
        self.assertEqual(uv.redirect_to, '/http/anchor/')
        self.assertEqual(uv.type, 'internal')

    @patch("linkcheck.models.TOLERATE_BROKEN_ANCHOR", False)
    def test_redirect_broken_internal_anchor(self):
        uv = Url(url="/http/redirect_to_anchor/#broken-anchor")
        uv.check_url()
        self.assertEqual(uv.message, "Working temporary redirect, broken internal hash anchor")
        self.assertEqual(uv.get_message, 'Working temporary redirect')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, False)
        self.assertEqual(uv.anchor_message, 'Broken anchor')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, '')
        self.assertEqual(uv.get_status_code_display(), '302 Found')
        self.assertEqual(uv.get_redirect_status_code_display(), '200 OK')
        self.assertEqual(uv.redirect_to, '/http/anchor/')
        self.assertEqual(uv.type, 'internal')

    def test_redirect_broken_internal_anchor_tolerated(self):
        uv = Url(url="/http/redirect_to_anchor/#broken-anchor")
        uv.check_url()
        self.assertEqual(uv.message, "Working temporary redirect, broken internal hash anchor")
        self.assertEqual(uv.get_message, 'Working temporary redirect')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, 'Broken anchor')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, '')
        self.assertEqual(uv.get_status_code_display(), '302 Found')
        self.assertEqual(uv.get_redirect_status_code_display(), '200 OK')
        self.assertEqual(uv.redirect_to, '/http/anchor/')
        self.assertEqual(uv.type, 'internal')

    def test_internal_check_working_redirect(self):
        uv = Url(url="/admin/linkcheck")
        uv.check_url()
        self.assertEqual(uv.message, "Working temporary redirect")
        self.assertEqual(uv.get_message, 'Working temporary redirect')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, '')
        self.assertEqual(uv.get_status_code_display(), '302 Found')
        self.assertEqual(uv.get_redirect_status_code_display(), '200 OK')
        self.assertEqual(uv.redirect_to, '/admin/login/?next=/admin/linkcheck')
        self.assertEqual(uv.type, 'internal')

    def test_internal_check_broken_redirect(self):
        uv = Url(url="/http/brokenredirect/")
        uv.check_url()
        self.assertEqual(uv.message, 'Broken temporary redirect')
        self.assertEqual(uv.get_message, 'Broken temporary redirect')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, False)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, '')
        self.assertEqual(uv.get_status_code_display(), '302 Found')
        self.assertEqual(uv.get_redirect_status_code_display(), '404 Not Found')
        self.assertEqual(uv.redirect_to, '/non-existent/')
        self.assertEqual(uv.type, 'internal')

    def test_internal_check_found(self):
        uv = Url(url="/public/")
        uv.check_url()
        self.assertEqual(uv.message, 'Working internal link')
        self.assertEqual(uv.get_message, 'Working internal link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, '')
        self.assertEqual(uv.get_status_code_display(), '200 OK')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'internal')

    def test_internal_check_with_protocol(self):
        # "localhost" is configured as SITE_DOMAIN in settings
        uv = Url(url="http://localhost/public/")
        uv.check_url()
        self.assertEqual(uv.message, 'Working internal link')
        self.assertEqual(uv.get_message, 'Working internal link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, '')
        self.assertEqual(uv.get_status_code_display(), '200 OK')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'internal')

    def test_internal_check_broken_internal_link(self):
        uv = Url(url="/broken/internal/link")
        uv.check_url()
        self.assertEqual(uv.message, 'Broken internal link')
        self.assertEqual(uv.get_message, 'Broken internal link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, False)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, '')
        self.assertEqual(uv.get_status_code_display(), '404 Not Found')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'internal')

    def test_internal_check_invalid_url(self):
        uv = Url(url="invalid/url")
        uv.check_url()
        self.assertEqual(uv.message, 'Invalid URL')
        self.assertEqual(uv.get_message, 'Invalid URL')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, False)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, '')
        self.assertEqual(uv.get_status_code_display(), None)
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'invalid')


class InternalMediaCheckTestCase(TestCase):
    def setUp(self):
        self.old_media_root = settings.MEDIA_ROOT
        settings.MEDIA_ROOT = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'media')

    def tearDown(self):
        settings.MEDIA_ROOT = self.old_media_root

    def test_internal_check_media_missing(self):
        uv = Url(url="/media/not_found")
        uv.check_url()
        self.assertEqual(uv.message, 'Missing Document')
        self.assertEqual(uv.get_message, 'Missing file')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, False)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, '')
        self.assertEqual(uv.get_status_code_display(), None)
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'file')

    def test_internal_check_media_found(self):
        uv = Url(url="/media/found")
        uv.check_url()
        self.assertEqual(uv.message, 'Working file link')
        self.assertEqual(uv.get_message, 'Working file link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, '')
        self.assertEqual(uv.get_status_code_display(), None)
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'file')

    def test_internal_check_media_utf8(self):
        media_file = os.path.join(os.path.dirname(__file__), 'media', 'rückmeldung')
        open(media_file, 'a').close()
        self.addCleanup(os.remove, media_file)
        uv = Url(url="/media/r%C3%BCckmeldung")
        uv.check_url()
        self.assertEqual(uv.message, 'Working file link')
        self.assertEqual(uv.get_message, 'Working file link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, '')
        self.assertEqual(uv.get_status_code_display(), None)
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'file')
        # Also when the url is not encoded
        uv = Url(url="/media/rückmeldung")
        uv.check_url()
        self.assertEqual(uv.message, 'Working file link')
        self.assertEqual(uv.get_message, 'Working file link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, '')
        self.assertEqual(uv.get_status_code_display(), None)
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'file')


@override_settings(SITE_DOMAIN='example.com')
class ExternalCheckTestCase(LiveServerTestCase):

    def setUp(self):
        urllib3.disable_warnings()

    def test_external_check_200(self):
        uv = Url(url=f"{self.live_server_url}/http/200/")
        uv.check_url()
        self.assertEqual(uv.message, '200 OK')
        self.assertEqual(uv.get_message, 'Working external link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'Insecure link')
        self.assertEqual(uv.get_status_code_display(), '200 OK')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    def test_external_check_200_valid_cert(self):
        uv = Url(url='https://rsa4096.badssl.com/')
        uv.check_url()
        self.assertEqual(uv.message, '200 OK')
        self.assertEqual(uv.get_message, 'Working external link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, True)
        self.assertEqual(uv.ssl_message, 'Valid SSL certificate')
        self.assertEqual(uv.get_status_code_display(), '200 OK')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    def test_external_check_200_missing_cert(self):
        uv = Url(url=f"{self.live_server_url.replace('http://', 'https://')}/http/200/")
        uv.check_url()
        self.assertEqual(uv.message, 'SSL Error: wrong version number')
        self.assertEqual(uv.get_message, 'SSL Error: wrong version number')
        self.assertEqual(uv.error_message, 'SSL Error: wrong version number')
        self.assertEqual(uv.status, False)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, False)
        self.assertEqual(uv.ssl_message, 'Broken SSL certificate')
        self.assertEqual(uv.get_status_code_display(), None)
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    def test_external_check_200_incomplete_cert(self):
        uv = Url(url="https://incomplete-chain.badssl.com/")
        uv.check_url()
        self.assertEqual(uv.message, '200 OK, SSL certificate could not be verified')
        self.assertEqual(uv.get_message, 'Working external link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'SSL certificate could not be checked')
        self.assertEqual(uv.get_status_code_display(), '200 OK')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    def test_external_check_200_broken_anchor_incomplete_cert(self):
        uv = Url(url="https://incomplete-chain.badssl.com/#broken")
        uv.check_url()
        self.assertEqual(uv.message, '200 OK, broken external hash anchor, SSL certificate could not be verified')
        self.assertEqual(uv.get_message, 'Working external link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, 'Broken anchor')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'SSL certificate could not be checked')
        self.assertEqual(uv.get_status_code_display(), '200 OK')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    def test_external_check_404_incomplete_cert(self):
        uv = Url(url="https://incomplete-chain.badssl.com/404")
        uv.check_url()
        self.assertEqual(uv.message, '404 Not Found, SSL certificate could not be verified')
        self.assertEqual(uv.get_message, 'Broken external link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, False)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'SSL certificate could not be checked')
        self.assertEqual(uv.get_status_code_display(), '404 Not Found')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    @requests_mock.Mocker()
    def test_external_check_unreachable(self, mocker):
        exc = ConnectionError(
            "HTTPSConnectionPool(host='name-resolution-error.example.com', port=443): Max retries exceeded with url: / "
            "(Caused by NameResolutionError(\"<urllib3.connection.HTTPSConnection object at 0xdeadbeef>: "
            "Failed to resolve 'name-resolution-error.example.com' ([Errno -2] Name or service not known)\"))"
        )
        mocked_url = 'https://name-resolution-error.example.com/'
        mocker.register_uri('HEAD', mocked_url, exc=exc),
        uv = Url(url=mocked_url)
        uv.check_url()
        formatted_message = (
            "Name Resolution Error: Failed to resolve 'name-resolution-error.example.com' "
            "([Errno -2] Name or service not known)"
        )
        self.assertEqual(uv.message, formatted_message)
        self.assertEqual(uv.get_message, formatted_message)
        self.assertEqual(uv.error_message, formatted_message)
        self.assertEqual(uv.status, False)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'SSL certificate could not be checked')
        self.assertEqual(uv.get_status_code_display(), None)
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    def test_external_check_200_utf8(self):
        uv = Url(url=f"{self.live_server_url}/http/200/r%C3%BCckmeldung/")
        uv.check_url()
        self.assertEqual(uv.message, '200 OK')
        self.assertEqual(uv.get_message, 'Working external link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'Insecure link')
        self.assertEqual(uv.get_status_code_display(), '200 OK')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    def test_external_check_200_utf8_not_encoded(self):
        uv = Url(url=f"{self.live_server_url}/http/200/rückmeldung/")
        uv.check_url()
        self.assertEqual(uv.message, '200 OK')
        self.assertEqual(uv.get_message, 'Working external link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'Insecure link')
        self.assertEqual(uv.get_status_code_display(), '200 OK')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    @requests_mock.Mocker()
    def test_external_check_200_utf8_domain(self, mocker):
        mocker.register_uri('HEAD', 'https://xn--utf8-test--z5a0txc.example.com/', reason='OK'),
        uv = Url(url='https://utf8-test-äüö.example.com/')
        uv.check_url()
        self.assertEqual(uv.message, '200 OK')
        self.assertEqual(uv.get_message, 'Working external link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, True)
        self.assertEqual(uv.ssl_message, 'Valid SSL certificate')
        self.assertEqual(uv.get_status_code_display(), '200 OK')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    @requests_mock.Mocker()
    def test_external_check_200_punycode_domain(self, mocker):
        punycode_domain = 'https://xn--utf8-test--z5a0txc.example.com/'
        mocker.register_uri('HEAD', punycode_domain, reason='OK'),
        uv = Url(url=punycode_domain)
        uv.check_url()
        self.assertEqual(uv.message, '200 OK')
        self.assertEqual(uv.get_message, 'Working external link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, True)
        self.assertEqual(uv.ssl_message, 'Valid SSL certificate')
        self.assertEqual(uv.get_status_code_display(), '200 OK')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    def test_external_check_301(self):
        uv = Url(url=f"{self.live_server_url}/http/301/")
        uv.check_url()
        self.assertEqual(uv.message, '301 Moved Permanently')
        self.assertEqual(uv.get_message, 'Broken permanent redirect')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, False)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'Insecure link')
        self.assertEqual(uv.get_status_code_display(), '301 Moved Permanently')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    def test_external_check_301_followed(self):
        uv = Url(url=f"{self.live_server_url}/http/redirect/301/")
        uv.check_url()
        self.assertEqual(uv.message, '301 Moved Permanently')
        self.assertEqual(uv.get_message, 'Working permanent redirect')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'Insecure link')
        self.assertEqual(uv.get_status_code_display(), '301 Moved Permanently')
        self.assertEqual(uv.get_redirect_status_code_display(), '200 OK')
        self.assertEqual(uv.redirect_to, f'{self.live_server_url}/http/200/')
        self.assertEqual(uv.type, 'external')

    def test_external_check_302_followed(self):
        uv = Url(url=f"{self.live_server_url}/http/redirect/302/")
        uv.check_url()
        self.assertEqual(uv.message, '302 Found')
        self.assertEqual(uv.get_message, 'Working temporary redirect')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'Insecure link')
        self.assertEqual(uv.get_status_code_display(), '302 Found')
        self.assertEqual(uv.get_redirect_status_code_display(), '200 OK')
        self.assertEqual(uv.redirect_to, f'{self.live_server_url}/http/200/')
        self.assertEqual(uv.type, 'external')

    def test_external_check_404(self):
        uv = Url(url=f"{self.live_server_url}/whatever/")
        uv.check_url()
        self.assertEqual(uv.message, '404 Not Found')
        self.assertEqual(uv.get_message, 'Broken external link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, False)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'Insecure link')
        self.assertEqual(uv.get_status_code_display(), '404 Not Found')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    def test_external_check_redirect_final_404(self):
        uv = Url(url=f"{self.live_server_url}/http/redirect_to_404/")
        uv.check_url()
        self.assertEqual(uv.message, '404 Not Found')
        self.assertEqual(uv.get_message, 'Broken permanent redirect')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, False)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'Insecure link')
        self.assertEqual(uv.get_status_code_display(), '301 Moved Permanently')
        self.assertEqual(uv.get_redirect_status_code_display(), '404 Not Found')
        self.assertEqual(uv.redirect_to, f'{self.live_server_url}/http/404/')
        self.assertEqual(uv.type, 'external')

    def test_external_check_get_only_405(self):
        # An URL that allows GET but not HEAD, linkcheck should fallback on GET.
        uv = Url(url=f"{self.live_server_url}/http/getonly/405/")
        uv.check_url()
        self.assertEqual(uv.message, '200 OK')
        self.assertEqual(uv.get_message, 'Working external link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'Insecure link')
        self.assertEqual(uv.get_status_code_display(), '200 OK')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    def test_external_check_get_only_400(self):
        uv = Url(url=f"{self.live_server_url}/http/getonly/400/")
        uv.check_url()
        self.assertEqual(uv.message, '200 OK')
        self.assertEqual(uv.get_message, 'Working external link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'Insecure link')
        self.assertEqual(uv.get_status_code_display(), '200 OK')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    def test_external_check_blocked_user_agent(self):
        uv = Url(url=f"{self.live_server_url}/http/block-user-agent/")
        uv.check_url()
        self.assertEqual(uv.message, '200 OK')
        self.assertEqual(uv.get_message, 'Working external link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'Insecure link')
        self.assertEqual(uv.get_status_code_display(), '200 OK')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    def test_external_check_blocked_user_agent_blocked_head(self):
        uv = Url(url=f"{self.live_server_url}/http/block-user-agent/block-head/")
        uv.check_url()
        self.assertEqual(uv.message, '200 OK')
        self.assertEqual(uv.get_message, 'Working external link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'Insecure link')
        self.assertEqual(uv.get_status_code_display(), '200 OK')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    @patch(
        'linkcheck.models.PROXIES',
        {'http': 'http://proxy.example.com:8080'},
    )
    @requests_mock.Mocker()
    def test_external_proxy_request(self, mocker):
        mocker.register_uri('HEAD', 'http://test.com', reason='OK'),
        uv = Url(url='http://test.com')
        self.assertEqual(mocker.called, False)
        uv.check_url()
        self.assertEqual(mocker.called, True)
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.message, '200 OK')
        self.assertEqual(uv.type, 'external')
        last_request = mocker.last_request
        self.assertEqual(last_request.hostname, 'test.com')
        self.assertEqual(last_request.scheme, 'http')
        self.assertEqual(last_request.proxies, {'http': 'http://proxy.example.com:8080'})

    def test_external_check_timedout(self):
        uv = Url(url=f"{self.live_server_url}/timeout/")
        uv.check_url()
        self.assertEqual(uv.message, 'Other Error: The read operation timed out')
        self.assertEqual(uv.get_message, 'The read operation timed out')
        self.assertEqual(uv.error_message, 'The read operation timed out')
        self.assertEqual(uv.status, False)
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'Insecure link')
        self.assertEqual(uv.get_status_code_display(), None)
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    def test_external_check_rate_limit(self):
        uv = Url(url=f"{self.live_server_url}/http/429/")
        uv.check_url()
        self.assertEqual(uv.last_checked, None)
        self.assertEqual(uv.message, '429 Too Many Requests')
        self.assertEqual(uv.get_message, 'Broken external link')
        self.assertEqual(uv.status, False)
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.anchor_message, '')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'Insecure link')
        self.assertEqual(uv.get_status_code_display(), '429 Too Many Requests')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    def test_working_external_anchor(self):
        uv = Url(url=f"{self.live_server_url}/http/anchor/#anchor")
        uv.check_url()
        self.assertEqual(uv.message, "200 OK, working external hash anchor")
        self.assertEqual(uv.get_message, 'Working external link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, 'Working anchor')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'Insecure link')
        self.assertEqual(uv.get_status_code_display(), '200 OK')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    @patch("linkcheck.models.TOLERATE_BROKEN_ANCHOR", False)
    def test_broken_external_anchor(self):
        uv = Url(url=f"{self.live_server_url}/http/anchor/#broken-anchor")
        uv.check_url()
        self.assertEqual(uv.message, "200 OK, broken external hash anchor")
        self.assertEqual(uv.get_message, 'Working external link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, False)
        self.assertEqual(uv.anchor_message, 'Broken anchor')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'Insecure link')
        self.assertEqual(uv.get_status_code_display(), '200 OK')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    def test_broken_external_anchor_tolerated(self):
        uv = Url(url=f"{self.live_server_url}/http/anchor/#broken-anchor")
        uv.check_url()
        self.assertEqual(uv.message, "200 OK, broken external hash anchor")
        self.assertEqual(uv.get_message, 'Working external link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, 'Broken anchor')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'Insecure link')
        self.assertEqual(uv.get_status_code_display(), '200 OK')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    def test_redirect_working_external_anchor(self):
        uv = Url(url=f"{self.live_server_url}/http/redirect_to_anchor/#anchor")
        uv.check_url()
        self.assertEqual(uv.message, "302 Found, working external hash anchor")
        self.assertEqual(uv.get_message, 'Working temporary redirect')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, 'Working anchor')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'Insecure link')
        self.assertEqual(uv.get_status_code_display(), '302 Found')
        self.assertEqual(uv.get_redirect_status_code_display(), '200 OK')
        self.assertEqual(uv.redirect_to, f'{self.live_server_url}/http/anchor/')
        self.assertEqual(uv.type, 'external')

    @patch("linkcheck.models.TOLERATE_BROKEN_ANCHOR", False)
    def test_redirect_broken_external_anchor(self):
        uv = Url(url=f"{self.live_server_url}/http/redirect_to_anchor/#broken-anchor")
        uv.check_url()
        self.assertEqual(uv.message, "302 Found, broken external hash anchor")
        self.assertEqual(uv.get_message, 'Working temporary redirect')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, False)
        self.assertEqual(uv.anchor_message, 'Broken anchor')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'Insecure link')
        self.assertEqual(uv.get_status_code_display(), '302 Found')
        self.assertEqual(uv.get_redirect_status_code_display(), '200 OK')
        self.assertEqual(uv.redirect_to, f'{self.live_server_url}/http/anchor/')
        self.assertEqual(uv.type, 'external')

    def test_redirect_broken_external_anchor_tolerated(self):
        uv = Url(url=f"{self.live_server_url}/http/redirect_to_anchor/#broken-anchor")
        uv.check_url()
        self.assertEqual(uv.message, "302 Found, broken external hash anchor")
        self.assertEqual(uv.get_message, 'Working temporary redirect')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, 'Broken anchor')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'Insecure link')
        self.assertEqual(uv.get_status_code_display(), '302 Found')
        self.assertEqual(uv.get_redirect_status_code_display(), '200 OK')
        self.assertEqual(uv.redirect_to, f'{self.live_server_url}/http/anchor/')
        self.assertEqual(uv.type, 'external')

    def test_video_with_time_anchor(self):
        uv = Url(url=f"{self.live_server_url}/static-files/video.mp4#t=2.0")
        uv.check_url()
        self.assertEqual(uv.message, "200 OK")
        self.assertEqual(uv.get_message, 'Working external link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.anchor_message, 'Anchor could not be checked')
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'Insecure link')
        self.assertEqual(uv.get_status_code_display(), '200 OK')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')

    def test_forged_video_with_time_anchor(self):
        uv = Url(url=f"{self.live_server_url}/static-files/fake-video.mp4#<t=2.0")
        uv.check_url()
        if sys.version_info >= (3, 13):
            # Following fixes for https://github.com/python/cpython/issues/135661,
            # it is harder to make html parser fail.
            self.assertEqual(uv.message, "200 OK, broken external hash anchor")
            self.assertEqual(uv.anchor_message, "Broken anchor")
        else:
            self.assertEqual(uv.message, "200 OK, failed to parse HTML for anchor")
            self.assertEqual(uv.anchor_message, "Anchor could not be checked")
        self.assertEqual(uv.get_message, 'Working external link')
        self.assertEqual(uv.error_message, '')
        self.assertEqual(uv.status, True)
        self.assertEqual(uv.ssl_status, None)
        self.assertEqual(uv.ssl_message, 'Insecure link')
        self.assertEqual(uv.get_status_code_display(), '200 OK')
        self.assertEqual(uv.get_redirect_status_code_display(), None)
        self.assertEqual(uv.redirect_to, '')
        self.assertEqual(uv.type, 'external')


class ModelTestCase(TestCase):

    def test_str(self):
        Author.objects.create(name="John Smith", website="http://www.example.org/smith")
        self.assertEqual(
            str(Url.objects.first()),
            "http://www.example.org/smith",
        )
        self.assertEqual(
            str(Link.objects.first()),
            "http://www.example.org/smith (Author object (1))",
        )

    def test_repr(self):
        Author.objects.create(name="John Smith", website="http://www.example.org/smith")
        self.assertEqual(
            repr(Url.objects.first()),
            "<Url (id: 1, url: http://www.example.org/smith)>",
        )
        self.assertEqual(
            repr(Link.objects.first()),
            (
                "<Link (id: 1, url: <Url (id: 1, url: http://www.example.org/smith)>, "
                "source: <Author: Author object (1)>)>"
            ),
        )


class ChecklinksTestCase(TestCase):

    def test_checklinks_command(self):
        Book.objects.create(title='My Title', description="""
            Here's an external link: <a href="http://www.example.org">External</a>,
            an internal link: <a href="/public/">Internal</a>,
            and an image: <img src="http://www.example.org/logo.png" alt="logo">""")

        out = StringIO()
        call_command('checklinks', stdout=out)
        self.assertEqual(
            out.getvalue(),
            "Checking all links that haven't been tested for 10080 minutes.\n"
            "1 internal URLs and 0 external URLs have been checked.\n"
        )

        yesterday = datetime.now() - timedelta(days=1)
        Url.objects.all().update(last_checked=yesterday)
        out = StringIO()
        call_command('checklinks', externalinterval=20, stdout=out)
        self.assertEqual(
            out.getvalue(),
            "Checking all links that haven't been tested for 20 minutes.\n"
            "1 internal URLs and 2 external URLs have been checked.\n"
        )

        Url.objects.all().update(last_checked=yesterday)
        out = StringIO()
        call_command('checklinks', externalinterval=20, limit=1, stdout=out)
        self.assertEqual(
            out.getvalue(),
            "Checking all links that haven't been tested for 20 minutes.\n"
            "Will run maximum of 1 checks this run.\n"
            "1 internal URLs and 1 external URLs have been checked.\n"
        )


class FindingLinksTestCase(TestCase):
    def test_found_links(self):
        self.assertEqual(Url.objects.all().count(), 0)
        Book.objects.create(title='My Title', description="""
            Here's a link: <a href="http://www.example.org">Example</a>,
            and an image: <img src="http://www.example.org/logo.png" alt="logo">""")
        self.assertEqual(Url.objects.all().count(), 2)
        self.assertQuerySetEqual(
            Url.objects.all().order_by('url'),
            ["http://www.example.org", "http://www.example.org/logo.png"],
            transform=lambda obj: obj.url
        )

    def test_urls_exceeding_max_length(self):
        self.assertEqual(Url.objects.all().count(), 0)
        with self.assertLogs(logger="linkcheck", level="WARN") as cm:
            Book.objects.create(
                title="My Title",
                description=(
                    "Here's a link: <a href='http://www.example.org'>Example</a>, and here's a url exceeding "
                    f"the max length: <img src='http://www.example.org/{MAX_URL_LENGTH * 'X'}' alt='logo'>"
                ),
            )
        # We skip urls which are too long because we can't store them in the database
        self.assertIn(
            (
                "WARNING:linkcheck.listeners:URL exceeding max length will be skipped: "
                f"http://www.example.org/{MAX_URL_LENGTH * 'X'}"
            ),
            cm.output,
        )
        self.assertEqual(Url.objects.all().count(), 1)

    def test_empty_url_field(self):
        """
        Test that URLField empty content is excluded depending on ignore_empty list.
        """
        all_linklists = apps.get_app_config('linkcheck').all_linklists
        all_linklists['Authors'].ignore_empty = ['website']
        try:
            Author.objects.create(name="William Shakespeare")
            Author.objects.create(name="John Smith", website="http://www.example.org/smith")
            self.assertEqual(Url.objects.all().count(), 1)
        finally:
            all_linklists['Authors'].ignore_empty = []
        Author.objects.create(name="Alphonse Daudet")
        # This time, the empty 'website' is extracted
        self.assertEqual(Url.objects.all().count(), 2)

    def test_findlinks_command(self):
        # Disable listeners to only check the management command
        with disable_listeners():
            Author.objects.create(name="John Smith", website="https://www.example.org/smith")
            self.assertEqual(
                findlinks(),
                "Updating all links...\n"
                "Urls: 1 created, 0 deleted, 0 unchanged\n"
                "Links: 1 created, 0 deleted, 0 unchanged\n"
            )
            Author.objects.create(name="John Doe", website="https://www.example.org/doe")
            Book.objects.create(
                title='My Title',
                description="My fav author: <a href='https://www.example.org/doe'>John Doe</a>"
            )
            self.assertEqual(
                findlinks(),
                "Updating all links...\n"
                "Urls: 1 created, 0 deleted, 1 unchanged\n"
                "Links: 2 created, 0 deleted, 1 unchanged\n"
            )
            Author.objects.get(name="John Doe").delete()
            self.assertEqual(
                findlinks(),
                "Updating all links...\n"
                "Urls: 0 created, 0 deleted, 2 unchanged\n"
                "Links: 0 created, 1 deleted, 2 unchanged\n"
            )
            Book.objects.first().delete()
            self.assertEqual(
                findlinks(),
                "Updating all links...\n"
                "Urls: 0 created, 1 deleted, 1 unchanged\n"
                "Links: 0 created, 1 deleted, 1 unchanged\n"
            )


class ManagementCommandTestCase(TestCase):

    def test_linkcheck_suggest_config(self):
        """
        Test that the config of uncovered models is correctly suggested
        """
        out, err = get_command_output('linkcheck_suggest_config')
        self.assertEqual(
            out,
            'All covered models:\n'
            '\x1b[36msampleapp.Book\x1b[0m, \x1b[36msampleapp.Page\x1b[0m\n\n'
            'Suggested config for model sampleapp.UncoveredModel:\n'
            '\x1b[36mfrom sampleapp.models import UncoveredModel\n\n'
            'class UncoveredModelLinklist(Linklist):\n'
            '    model = UncoveredModel\n\n'
            'linklists = {\n'
            '    "UncoveredModel": UncoveredModelLinklist,\n'
            '}\n\x1b[0m\n'
        )
        self.assertEqual(err, '')

    def test_linkcheck_suggest_config_model(self):
        """
        Test that the config of given model is correctly printed
        """
        out, err = get_command_output('linkcheck_suggest_config', '--model', 'sampleapp.Author')
        self.assertEqual(
            out,
            'from sampleapp.models import Author\n\n'
            'class AuthorLinklist(Linklist):\n'
            '    model = Author\n\n'
            'linklists = {\n'
            '    "Author": AuthorLinklist,\n'
            '}\n'
        )
        self.assertEqual(err, '')

    def test_linkcheck_suggest_config_model_non_existing(self):
        """
        Test that the command raises an error when the model does not exist
        """
        with self.assertRaises(CommandError) as cm:
            get_command_output('linkcheck_suggest_config', '--model', 'non-existing')
        self.assertEqual(str(cm.exception), 'Model "non-existing" does not exist.')


class ObjectsUpdateTestCase(TestCase):
    def test_update_object(self):
        """
        Test that updating a broken URL in an object also updates the
        corresponding Link, and don't leak the old URL.
        """
        bad_url = "/broken/internal/link"
        good_url = "/public/"
        author = Author.objects.create(name="John Smith", website=bad_url)
        self.assertEqual(
            Link.objects.filter(ignore=False, url__status=False).count(),
            1
        )
        self.assertEqual(
            Link.objects.filter(ignore=False, url__status=True).count(),
            0
        )
        self.assertEqual(Url.objects.all().count(), 1)
        self.assertEqual(Url.objects.all()[0].url, bad_url)
        # Fix the link
        author.website = good_url
        author.save()
        self.assertEqual(
            Link.objects.filter(ignore=False, url__status=False).count(),
            0
        )
        self.assertEqual(
            Link.objects.filter(ignore=False, url__status=True).count(),
            1
        )
        self.assertEqual(Url.objects.all().count(), 1)
        self.assertEqual(Url.objects.all()[0].url, good_url)


class RegisteringTests(TestCase):
    good_url = "/public/"

    def test_unregister(self):
        self.assertEqual(Link.objects.count(), 0)
        unregister_listeners()
        Author.objects.create(name="John Smith", website=self.good_url)
        self.assertEqual(Link.objects.count(), 0)
        register_listeners()
        Author.objects.create(name="Jill Smith", website=self.good_url)
        self.assertEqual(Link.objects.count(), 1)

    def test_disable_listeners(self):
        self.assertEqual(Link.objects.count(), 0)
        with disable_listeners():
            Author.objects.create(name="John Smith", website=self.good_url)
        self.assertEqual(Link.objects.count(), 0)
        Author.objects.create(name="Jill Smith", website=self.good_url)
        self.assertEqual(Link.objects.count(), 1)

    def test_enable_listeners(self):
        self.assertEqual(Link.objects.count(), 0)
        unregister_listeners()
        with enable_listeners():
            Author.objects.create(name="John Smith", website=self.good_url)
        self.assertEqual(Link.objects.count(), 1)
        Author.objects.create(name="Jill Smith", website=self.good_url)
        self.assertEqual(Link.objects.count(), 1)
        register_listeners()


class QueueTests(TestCase):
    def test_queue_handling_continue_on_task_crash(self):
        assert tasks_queue.empty() is True

        def raising():
            raise RuntimeError("Failing task")

        def passing():
            pass

        for func in (raising, passing):
            tasks_queue.put({
                'target': func,
                'args': (),
                'kwargs': {},
            })
        with self.assertLogs() as cm:
            linkcheck_worker(block=False)
        self.assertEqual(
            cm.output[0].split('\n')[0],
            'ERROR:linkcheck.listeners:RuntimeError while running raising with '
            'args=() and kwargs={}: Failing task'
        )


class ViewTestCase(TestCase):
    def setUp(self):
        self.user = User.objects.create_superuser('admin', 'admin@example.org', 'password')

    def test_display_url(self):
        Book.objects.create(
            title='My Title', description="Here's a link: <a href='http://www.example.org'>Example</a>"
        )
        Author.objects.create(name="John Smith", website="http://www.example.org#john")
        self.assertEqual(Link.objects.count(), 2)
        self.assertEqual(
            set([link.display_url for link in Link.objects.all()]),
            set(["http://www.example.org", "http://www.example.org#john"]),
        )

    def test_report_view(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('linkcheck_report'))
        self.assertContains(response, "<h1>Link Checker</h1>")

    def test_report_ignore_unignore(self):
        Author.objects.create(name="John Smith", website="http://www.example.org/john")
        self.client.force_login(self.user)
        link = Link.objects.first()
        self.assertFalse(link.ignore)
        response = self.client.post(
            reverse('linkcheck_report') + f"?ignore={link.pk}",
            HTTP_X_REQUESTED_WITH='XMLHttpRequest'
        )
        self.assertEqual(response.json(), {'link': link.pk})
        link.refresh_from_db()
        self.assertTrue(link.ignore)
        response = self.client.post(
            reverse('linkcheck_report') + f"?unignore={link.pk}",
            HTTP_X_REQUESTED_WITH='XMLHttpRequest'
        )
        self.assertEqual(response.json(), {'link': link.pk})
        link.refresh_from_db()
        self.assertFalse(link.ignore)

    def test_report_recheck(self):
        Author.objects.create(name="John Smith", website="http://www.example.org/john")
        self.client.force_login(self.user)
        link = Link.objects.first()
        response = self.client.post(
            reverse('linkcheck_report') + f"?recheck={link.pk}",
            HTTP_X_REQUESTED_WITH='XMLHttpRequest'
        )
        self.assertEqual(response.json(), {
            'colour': 'red',
            'links': [link.pk],
            'message': '404 Not Found',
        })


class GetJqueryMinJsTestCase(TestCase):
    def test(self):
        self.assertEqual(
            'admin/js/vendor/jquery/jquery.min.js', get_jquery_min_js()
        )


class FixtureTestCase(TestCase):
    fixtures = ['linkcheck/tests/sampleapp/fixture.json']

    def test_fixture(self):
        self.assertEqual(Book.objects.count(), 1)
        self.assertEqual(Page.objects.count(), 1)


class FilterCallableTestCase(TestCase):
    def test_filter_callable(self):
        all_linklists = apps.get_app_config('linkcheck').all_linklists
        all_linklists['Journals'].html_fields = []
        Journal.objects.create(title='My Title', description="""
            My description <a href="http://www.example.org">Example</a>""")
        Journal.objects.create(title='My Title', version=1, description="""
            My new description <a href="http://www.example.org">Example</a>""")
        all_linklists['Journals'].html_fields = ['description']
        # assert there are two versions of the same journal
        self.assertEqual(Journal.objects.count(), 2)
        # assert command just finds the latest version of same journals
        self.assertEqual(
            findlinks(),
            "Updating all links...\n"
            "Urls: 1 created, 0 deleted, 0 unchanged\n"
            "Links: 1 created, 0 deleted, 0 unchanged\n"
        )


def get_command_output(command, *args, **kwargs):
    """
    Helper function for running a management command and checking its output
    """
    out = StringIO()
    err = StringIO()
    call_command(command, *args, stdout=out, stderr=err, **kwargs)
    return out.getvalue(), err.getvalue()


def findlinks():
    """
    Helper function for running the findlinks command and checking its output
    """
    return get_command_output('findlinks')[0]



from django.core.management.base import BaseCommand

from linkcheck.utils import unignore


class Command(BaseCommand):

    help = "Updates the `ignore` status of all links to `False`"

    def execute(self, *args, **options):
        print("Unignoring all links")
        unignore()



from django.db import models


class Book(models.Model):
    title = models.CharField(max_length=50)
    description = models.TextField()

    def get_absolute_url(self):
        return f"/book/{self.id}/"


class Page(models.Model):
    book = models.ForeignKey(Book, on_delete=models.CASCADE)

    def get_absolute_url(self):
        return f"/book/{self.book.id}/{self.id}"


class Author(models.Model):
    # This model has purposefully no get_absolute_url
    name = models.CharField(max_length=50)
    website = models.URLField(blank=True)


class Journal(models.Model):
    title = models.CharField(max_length=50)
    description = models.TextField()
    version = models.PositiveIntegerField(default=0)


class UncoveredModel(models.Model):
    book = models.ForeignKey(Book, on_delete=models.CASCADE)
    website = models.URLField(blank=True)

    def get_absolute_url(self):
        return f'/uncovered/{self.id}'



from django.db.models import OuterRef, Subquery

from linkcheck import Linklist
from linkcheck.tests.sampleapp.models import Author, Book, Journal, Page


class BookLinklist(Linklist):
    """ Class to let linkcheck app discover fields containing links """
    model = Book
    object_filter = {}
    html_fields = ['description']


class PageLinklist(Linklist):
    """ Class to let linkcheck app discover fields containing links """
    model = Page


class AuthorLinklist(Linklist):
    """ Class to let linkcheck app discover fields containing links """
    model = Author
    object_filter = {}
    url_fields = ['website']


class JournalLinklist(Linklist):
    """ Class to let linkcheck app discover fields containing links """
    model = Journal
    html_fields = ['description']

    @classmethod
    def filter_callable(cls, objects):
        latest = Journal.objects.filter(title=OuterRef('title')).order_by('-version')
        return objects.filter(version=Subquery(latest.values('version')[:1]))


linklists = {
    'Books': BookLinklist,
    'Pages': PageLinklist,
    'Authors': AuthorLinklist,
    'Journals': JournalLinklist,
}



from django.db import migrations, models

from linkcheck.models import STATUS_CODE_CHOICES


class Migration(migrations.Migration):

    dependencies = [
        ("linkcheck", "0006_url_add_status_code"),
    ]

    operations = [
        migrations.AddField(
            model_name="url",
            name="redirect_status_code",
            field=models.IntegerField(
                choices=STATUS_CODE_CHOICES,
                null=True,
            ),
        ),
    ]



import logging
import time

from linkcheck.models import Link, Url

from . import update_lock
from .linkcheck_settings import MAX_URL_LENGTH

logger = logging.getLogger(__name__)


def do_check_instance_links(sender, instance, linklist_cls, wait=False):
    # On some installations, this wait time might be enough for the
    # thread transaction to account for the object change (GH #41).
    # A candidate for the future post_commit signal.

    if wait:
        time.sleep(0.1)
    with update_lock:
        content_type = linklist_cls.content_type()
        new_links = []
        old_links = Link.objects.filter(content_type=content_type, object_id=instance.pk)

        linklists = linklist_cls().get_linklist(extra_filter={'pk': instance.pk})

        if not linklists:
            # This object is no longer watched by linkcheck according to object_filter
            links = []
        else:
            linklist = linklists[0]
            links = linklist['urls']+linklist['images']

        for link in links:
            # url structure = (field, link text, url)
            url = link[2]
            if url.startswith('#'):
                url = instance.get_absolute_url() + url

            if len(url) > MAX_URL_LENGTH:
                # We cannot handle url longer than MAX_URL_LENGTH at the moment
                logger.warning('URL exceeding max length will be skipped: %s', url)
                continue

            u, created = Url.objects.get_or_create(url=url)
            l, created = Link.objects.get_or_create(
                url=u, field=link[0], text=link[1], content_type=content_type, object_id=instance.pk
            )
            new_links.append(l.id)
            u.check_url()

        gone_links = old_links.exclude(id__in=new_links)
        gone_links.delete()


def do_instance_post_save(sender, instance, linklist_cls, **kwargs):
    current_url = instance.get_absolute_url()
    previous_url = getattr(instance, '__previous_url', None)
    # We assume returning None from get_absolute_url means that this instance doesn't have a URL
    # Not sure if we should do the same for '' as this could refer to '/'
    if current_url is not None and current_url != previous_url:
        linklist_cls = sender._linklist
        active = linklist_cls.objects().filter(pk=instance.pk).count()

        if kwargs['created'] or (not active):
            new_urls = Url.objects.filter(url__startswith=current_url)
        else:
            new_urls = Url.objects.filter(status=False).filter(url__startswith=current_url)

        if new_urls:
            for url in new_urls:
                url.check_url()


def do_instance_pre_delete(sender, instance, **kwargs):
    instance.linkcheck_deleting = True
    deleted_url = instance.get_absolute_url()
    if deleted_url:
        old_urls = Url.objects.filter(url__startswith=deleted_url).exclude(status=False)
        if old_urls:
            old_urls.update(status=False, message='Broken internal link')



import time

from django.core.exceptions import PermissionDenied
from django.http import (
    HttpResponse,
    HttpResponsePermanentRedirect,
    HttpResponseRedirect,
)


def http_response(request, code):
    return HttpResponse("", status=int(code))


def http_response_get_only(request, code):
    status = int(code) if request.method == 'HEAD' else 200
    return HttpResponse("", status=status)


def http_block_user_agent(request, block_head=False):
    if block_head and request.method == 'HEAD':
        return HttpResponse('', status=405)
    if 'Linkchecker' in request.headers.get('User-Agent', ''):
        raise PermissionDenied()
    return HttpResponse('')


def http_redirect(request, code):
    return HttpResponseRedirect("/http/200/", status=int(code))


def http_redirect_to_404(request):
    return HttpResponsePermanentRedirect("/http/404/")


def timeout(request):
    time.sleep(2)
    return HttpResponse("")


def http_response_with_anchor(request):
    return HttpResponse("<html><body><h1 id='anchor'>Anchor</h1></body></html>")


def http_redirect_to_anchor(request):
    return HttpResponseRedirect("/http/anchor/")


def static_video(request):
    return HttpResponse(b'', content_type='video/mp4')


def static_video_forged_content_type(request):
    return HttpResponse(b'<![x02\x00\xa0\xcc', content_type='text/html')



import logging
from datetime import timedelta

from django.apps import apps
from django.db import models
from django.test.client import ClientHandler
from django.utils import timezone

from .linkcheck_settings import (
    HTML_FIELD_CLASSES,
    IMAGE_FIELD_CLASSES,
    MAX_URL_LENGTH,
    URL_FIELD_CLASSES,
)
from .models import Link, Url

logger = logging.getLogger(__name__)


class LinkCheckHandler(ClientHandler):
    # Customize the ClientHandler to allow us removing some middlewares

    def load_middleware(self):
        self.ignore_keywords = ["reversion.middleware", "MaintenanceModeMiddleware", "raven_compat"]
        super().load_middleware()
        new_request_middleware = []

        ####################################################
        # _request_middleware (is removed in newer django) #
        ####################################################
        if getattr(self, "_request_middleware", None):
            for method in self._request_middleware:
                ignored = False
                for keyword in self.ignore_keywords:
                    if method.__str__().count(keyword):
                        ignored = True
                        break
                if not ignored:
                    new_request_middleware.append(method)
            self._request_middleware = new_request_middleware

        ####################
        # _view_middleware #
        ####################
        new_view_middleware = []
        for method in self._view_middleware:
            ignored = False
            for keyword in self.ignore_keywords:
                if method.__str__().count(keyword):
                    ignored = True
                    break
            if not ignored:
                new_view_middleware.append(method)
        self._view_middleware = new_view_middleware

        ##########################
        # _response_middleware## #
        ##########################
        if getattr(self, "_response_middleware", None):
            new_response_middleware = []
            for method in self._response_middleware:
                ignored = False
                for keyword in self.ignore_keywords:
                    if method.__str__().count(keyword):
                        ignored = True
                        break
                if not ignored:
                    new_response_middleware.append(method)
            self._response_middleware = new_response_middleware

        #################################
        # _template_response_middleware #
        #################################
        if getattr(self, "_template_response_middleware", None):
            new_template_response_middleware = []
            for method in self._template_response_middleware:
                ignored = False
                for keyword in self.ignore_keywords:
                    if method.__str__().count(keyword):
                        ignored = True
                        break
                if not ignored:
                    new_template_response_middleware.append(method)
            self._template_response_middleware = new_template_response_middleware

        #########################
        # _exception_middleware #
        #########################
        new_exception_middleware = []
        for method in self._exception_middleware:
            ignored = False
            for keyword in self.ignore_keywords:
                if method.__str__().count(keyword):
                    ignored = True
                    break
            if not ignored:
                new_exception_middleware.append(method)
        self._exception_middleware = new_exception_middleware


def check_links(external_recheck_interval=10080, limit=-1, check_internal=True, check_external=True):
    """
    Return the number of links effectively checked.
    """

    urls = Url.objects.all()

    # An optimization for when check_internal is False
    if not check_internal:
        recheck_datetime = timezone.now() - timedelta(minutes=external_recheck_interval)
        urls = urls.exclude(last_checked__gt=recheck_datetime)

    check_count = 0
    for u in urls:
        status = u.check_url(check_internal=check_internal, check_external=check_external)
        check_count += 1 if status is not None else 0
        if -1 < limit <= check_count:
            break

    return check_count


def update_urls(urls, content_type, object_id):
    # Structure of urls param is [(field, link text, url), ... ]

    urls_created = links_created = 0
    new_url_ids = set()
    new_link_ids = set()

    for field, link_text, url in urls:
        if url is not None and url.startswith("#"):
            instance = content_type.get_object_for_this_type(id=object_id)
            url = instance.get_absolute_url() + url

        if len(url) > MAX_URL_LENGTH:
            # We cannot handle url longer than MAX_URL_LENGTH at the moment
            logger.warning("URL exceeding max length will be skipped: %s", url)
            continue

        url, url_created = Url.objects.get_or_create(url=url)

        link, link_created = Link.objects.get_or_create(
            url=url,
            field=field,
            text=link_text,
            content_type=content_type,
            object_id=object_id,
        )

        # Keep track of how many objects were created
        urls_created += url_created
        links_created += link_created

        # Keep track of object ids (no matter if created or existing)
        new_url_ids.add(url.id)
        new_link_ids.add(link.id)

    return {
        "urls": {
            "created": urls_created,
            "ids": new_url_ids,
        },
        "links": {
            "created": links_created,
            "ids": new_link_ids,
        },
    }


def find_all_links(linklists=None):
    if linklists is None:
        linklists = apps.get_app_config("linkcheck").all_linklists

    urls_created = links_created = 0
    new_url_ids = set()
    new_link_ids = set()

    urls_before = Url.objects.count()
    links_before = Link.objects.count()

    for linklist_name, linklist_cls in linklists.items():
        content_type = linklist_cls.content_type()
        linklists = linklist_cls().get_linklist()

        for linklist in linklists:
            object_id = linklist["object"].pk
            urls = linklist["urls"] + linklist["images"]
            if urls:
                new = update_urls(urls, content_type, object_id)

                urls_created += new["urls"]["created"]
                links_created += new["links"]["created"]

                new_url_ids.update(new["urls"]["ids"])
                new_link_ids.update(new["links"]["ids"])

    # Delete all urls and links which are no longer part of the link lists
    Url.objects.all().exclude(id__in=new_url_ids).delete()
    Link.objects.all().exclude(id__in=new_link_ids).delete()

    # Calculate diff
    urls_after = Url.objects.count()
    links_after = Link.objects.count()

    return {
        "urls": {
            "created": urls_created,
            "deleted": urls_before + urls_created - urls_after,
            "unchanged": urls_after - urls_created,
        },
        "links": {
            "created": links_created,
            "deleted": links_before + links_created - links_after,
            "unchanged": links_after - links_created,
        },
    }


def unignore():
    Link.objects.update(ignore=False)


# Utilities for testing models coverage


def is_interesting_field(field):
    return is_url_field(field) or is_image_field(field) or is_html_field(field)


def is_url_field(field):
    return any(isinstance(field, cls) for cls in URL_FIELD_CLASSES)


def is_image_field(field):
    return any(isinstance(field, cls) for cls in IMAGE_FIELD_CLASSES)


def is_html_field(field):
    return any(isinstance(field, cls) for cls in HTML_FIELD_CLASSES)


def has_active_field(klass):
    return any(field.name == "active" and isinstance(field, models.BooleanField) for field in klass._meta.fields)


def get_ignore_empty_fields(klass):
    return [field for field in klass._meta.fields if is_interesting_field(field) and (field.blank or field.null)]


def get_type_fields(klass, the_type):
    check_funcs = {
        "html": is_html_field,
        "url": is_url_field,
        "image": is_image_field,
    }
    check_func = check_funcs[the_type]
    return [field for field in klass._meta.fields if check_func(field)]


def is_model_covered(klass):
    app = apps.get_app_config("linkcheck")
    return any(linklist[1].model == klass for linklist in app.all_linklists.items())


def format_config(meta, active_field, html_fields, image_fields, url_fields, ignore_empty_fields):
    config = f"from {meta.app_label}.models import {meta.object_name}\n\n"
    config += f"class {meta.object_name}Linklist(Linklist):\n"
    config += f"    model = {meta.object_name}\n"
    if html_fields:
        config += f"    html_fields = [{', '.join(map(str, html_fields))}]\n"
    if image_fields:
        config += f"    image_fields = [{', '.join(map(str, image_fields))}]\n"
    if url_fields:
        config += f"    url_fields = [{', '.join(map(str, url_fields))}]\n"
    if ignore_empty_fields:
        config += f"    ignore_empty = [{', '.join(map(str, ignore_empty_fields))}]\n"
    if active_field:
        config += '    object_filter = {"active": True}\n'
    config += f'\nlinklists = {{\n    "{meta.object_name}": {meta.object_name}Linklist,\n}}\n'
    return config


def get_suggested_linklist_config(klass):
    meta = klass._meta
    html_fields = get_type_fields(klass, "html")
    url_fields = get_type_fields(klass, "url")
    image_fields = get_type_fields(klass, "image")
    active_field = has_active_field(klass)
    ignore_empty_fields = get_ignore_empty_fields(klass)
    return format_config(
        **{
            "meta": meta,
            "html_fields": html_fields,
            "url_fields": url_fields,
            "image_fields": image_fields,
            "active_field": active_field,
            "ignore_empty_fields": ignore_empty_fields,
        }
    )


def get_coverage_data():
    """
    Check which models are covered by linkcheck
    This view assumes the key for link
    """
    covered = []
    uncovered = []
    for app in apps.get_app_configs():
        for model in app.get_models():
            should_append = False
            if getattr(model, "get_absolute_url", None):
                should_append = True
            else:
                for field in model._meta.fields:
                    if is_interesting_field(field):
                        should_append = True
                        break
            if should_append:
                if is_model_covered(model):
                    covered.append(f"{model._meta.app_label}.{model._meta.object_name}")
                else:
                    uncovered.append(
                        (
                            f"{model._meta.app_label}.{model._meta.object_name}",
                            get_suggested_linklist_config(model),
                        )
                    )

    return covered, uncovered



from django.core.management.base import BaseCommand

from linkcheck.utils import find_all_links


class Command(BaseCommand):

    help = (
        "Goes through all models registered with Linkcheck, records any new links found"
        "and removes all outdated links"
    )

    def handle(self, *args, **options):
        self.stdout.write("Updating all links...")
        return "\n".join(
            [
                f"{model.capitalize()}: {', '.join([f'{count} {label}' for label, count in data.items()])}"
                for model, data in find_all_links().items()
            ]
        )



"""Integrate with django-filebrowser if present."""
import logging
import os.path

from django.conf import settings
from django.contrib import messages
from django.utils.translation import gettext as _
from django.utils.translation import ngettext

try:
    from filebrowser.settings import DIRECTORY
    from filebrowser.signals import (
        filebrowser_post_delete,
        filebrowser_post_rename,
        filebrowser_post_upload,
    )
    FILEBROWSER_PRESENT = True
except ImportError:
    FILEBROWSER_PRESENT = False

from linkcheck.models import Url

logger = logging.getLogger(__name__)


def get_relative_media_url():
    if settings.MEDIA_URL.startswith('http'):
        relative_media_url = ('/'+'/'.join(settings.MEDIA_URL.split('/')[3:]))[:-1]
    else:
        relative_media_url = settings.MEDIA_URL
    return relative_media_url


def handle_upload(sender, path=None, **kwargs):
    logger.debug('uploaded path %s with kwargs %r', path, kwargs)

    url = os.path.join(get_relative_media_url(), kwargs['file'].url)
    url_qs = Url.objects.filter(url=url).filter(status=False)
    count = url_qs.count()
    if count:
        url_qs.update(status=True, message="Working document link")
        msg = ngettext(
            "Uploading {} has corrected {} broken link.",
            "Uploading {} has corrected {} broken links.",
            count,
        ).format(url, count)
        messages.success(sender, '{}: {} {}'.format(
            _('Please note'),
            msg,
            _('See the Link Checker for more details.')
        ))


def handle_rename(sender, path=None, **kwargs):
    logger.debug('renamed path %s with kwargs %r', path, kwargs)

    def isdir(filename):
        if filename.count('.'):
            return False
        else:
            return True

    old_url = os.path.join(get_relative_media_url(), DIRECTORY, path)
    new_url = os.path.join(get_relative_media_url(), DIRECTORY, path.replace(kwargs['name'], kwargs['new_name']))
    # Renaming a file will cause it's urls to become invalid
    # Renaming a directory will cause the urls of all it's contents to become invalid
    old_url_qs = Url.objects.filter(url=old_url).filter(status=True)
    if isdir(kwargs['name']):
        old_url_qs = Url.objects.filter(url__startswith=old_url).filter(status=True)
    old_count = old_url_qs.count()
    if old_count:
        old_url_qs.update(status=False, message="Missing Document")
        msg = ngettext(
            "Renaming {} has caused {} link to break.",
            "Renaming {} has caused {} links to break.",
            old_count,
        ).format(old_url, old_count)
        messages.warning(sender, '{}: {} {}'.format(
            _('Warning'),
            msg,
            _('Please use the Link Checker to fix them.')
        ))

    # The new directory may fix some invalid links, so we also check for that
    if isdir(kwargs['new_name']):
        new_count = 0
        new_url_qs = Url.objects.filter(url__startswith=new_url).filter(status=False)
        for url in new_url_qs:
            if url.check_url():
                new_count += 1
    else:
        new_url_qs = Url.objects.filter(url=new_url).filter(status=False)
        new_count = new_url_qs.count()
        if new_count:
            new_url_qs.update(status=True, message='Working document link')
    if new_count:
        msg = ngettext(
            "Renaming {} has corrected {} broken link.",
            "Renaming {} has corrected {} broken links.",
            new_count,
        ).format(new_url, new_count)
        messages.success(sender, '{}: {} {}'.format(
            _('Please note'),
            msg,
            _('See the Link Checker for more details.')
        ))


def handle_delete(sender, path=None, **kwargs):
    logger.debug('deleted path %s with kwargs %r', path, kwargs)

    url = os.path.join(get_relative_media_url(), DIRECTORY, path)
    url_qs = Url.objects.filter(url=url).filter(status=True)
    count = url_qs.count()
    if count:
        url_qs.update(status=False, message="Missing Document")
        msg = ngettext(
            "Deleting {} has caused {} link to break.",
            "Deleting {} has caused {} links to break.",
            count,
        ).format(url, count)
        messages.warning(sender, '{}: {} {}'.format(
            _('Warning'),
            msg,
            _('Please use the Link Checker to fix them.')
        ))


def register_listeners():
    if FILEBROWSER_PRESENT:
        filebrowser_post_upload.connect(handle_upload)
        filebrowser_post_rename.connect(handle_rename)
        filebrowser_post_delete.connect(handle_delete)


def unregister_listeners():
    if FILEBROWSER_PRESENT:
        filebrowser_post_upload.disconnect(handle_upload)
        filebrowser_post_rename.disconnect(handle_rename)
        filebrowser_post_delete.disconnect(handle_delete)



from django import template

register = template.Library()


@register.filter
def get_verbose_name_plural(content_type):
    """
    Returns verbose_name_plural for a content type.
    """
    return content_type.model_class()._meta.verbose_name_plural.title()



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('linkcheck', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='url',
            name='redirect_to',
            field=models.CharField(default='', max_length=255),
        ),
    ]



import logging
import sys
from contextlib import contextmanager
from queue import Empty, LifoQueue
from threading import Thread

from django.apps import apps
from django.db.models import signals as model_signals

from linkcheck.models import Link, Url

from . import filebrowser
from .linkcheck_settings import LINKCHECK_IN_CELERY
from .worker_tasks import (
    do_check_instance_links,
    do_instance_post_save,
    do_instance_pre_delete,
)

if LINKCHECK_IN_CELERY:
    from .celery import do_check_link, do_post_save

logger = logging.getLogger(__name__)


tasks_queue = LifoQueue()
worker_running = False
tests_running = len(sys.argv) > 1 and sys.argv[1] == 'test' or sys.argv[0].endswith('runtests.py')


def linkcheck_worker(block=True):
    global worker_running  # noqa
    while tasks_queue.not_empty:
        try:
            task = tasks_queue.get(block=block)
        except Empty:
            break
        # An error in any task should not stop the worker from continuing with the queue
        try:
            task['target'](*task['args'], **task['kwargs'])
        except Exception as e:
            logger.exception(
                "%s while running %s with args=%r and kwargs=%r: %s",
                type(e).__name__,
                task['target'].__name__,
                task['args'],
                task['kwargs'],
                e
            )
        tasks_queue.task_done()
    worker_running = False


def start_worker():
    global worker_running  # noqa
    if worker_running is False:
        worker_running = True
        t = Thread(target=linkcheck_worker)
        t.daemon = True
        t.start()


def check_instance_links(sender, instance, **kwargs):
    """
    When an object is saved:
        new Link/Urls are created, checked

    When an object is modified:
        new link/urls are created, checked
        existing link/urls are checked
        Removed links are deleted
    """
    linklist_cls = sender._linklist

    # Don't run in a separate thread if we are running tests
    if tests_running:
        do_check_instance_links(sender, instance, linklist_cls)
    elif LINKCHECK_IN_CELERY:
        do_check_link.apply_async(
            kwargs={
                "app": sender._meta.app_label,
                "model": sender.__name__,
                "id": instance.id,
            },
        )
    else:
        tasks_queue.put({
            'target': do_check_instance_links,
            'args': (sender, instance, linklist_cls, True),
            'kwargs': {}
        })
        start_worker()


def delete_instance_links(sender, instance, **kwargs):
    """
    Delete all links belonging to a model instance when that instance is deleted
    """
    linklist_cls = sender._linklist
    content_type = linklist_cls.content_type()
    old_links = Link.objects.filter(content_type=content_type, object_id=instance.pk)
    old_links.delete()


def instance_pre_save(sender, instance, raw=False, **kwargs):
    if instance._state.adding or not instance.pk or raw:
        # Ignore unsaved instances or raw imports
        return
    current_url = instance.get_absolute_url()
    previous_url = sender.objects.get(pk=instance.pk).get_absolute_url()
    setattr(instance, '__previous_url', previous_url)
    if previous_url == current_url:
        return
    else:
        if previous_url is not None:
            old_urls = Url.objects.filter(url__startswith=previous_url)
            old_urls.update(status=False, message='Broken internal link')
        if current_url is not None:
            new_urls = Url.objects.filter(url__startswith=current_url)
            # Mark these urls' status as False, so that post_save will check them
            new_urls.update(status=False, message='Should be checked now!')


def instance_post_save(sender, instance, **kwargs):
    # Ignore raw imports
    if kwargs.get('raw'):
        return

    linklist_cls = sender._linklist

    if tests_running:
        do_instance_post_save(sender, instance, linklist_cls, **kwargs)
    elif LINKCHECK_IN_CELERY:
        do_post_save.apply_async(
            kwargs={
                "app": sender._meta.app_label,
                "model": sender.__name__,
                "id": instance.id,
            },
        )
    else:
        tasks_queue.put({
            'target': do_instance_post_save,
            'args': (sender, instance, linklist_cls),
            'kwargs': kwargs
        })
        start_worker()


def register_listeners():
    # 1. register listeners for the objects that contain Links
    for linklist_name, linklist_cls in apps.get_app_config('linkcheck').all_linklists.items():
        model_signals.post_save.connect(check_instance_links, sender=linklist_cls.model)
        model_signals.post_delete.connect(delete_instance_links, sender=linklist_cls.model)

        # 2. register listeners for the objects that are targets of Links,
        # only when get_absolute_url() is defined for the model
        if getattr(linklist_cls.model, 'get_absolute_url', None):
            model_signals.pre_save.connect(instance_pre_save, sender=linklist_cls.model)
            model_signals.post_save.connect(instance_post_save, sender=linklist_cls.model)
            model_signals.pre_delete.connect(do_instance_pre_delete, sender=linklist_cls.model)

    filebrowser.register_listeners()


def unregister_listeners():
    # 1. register listeners for the objects that contain Links
    for linklist_name, linklist_cls in apps.get_app_config('linkcheck').all_linklists.items():
        model_signals.post_save.disconnect(check_instance_links, sender=linklist_cls.model)
        model_signals.post_delete.disconnect(delete_instance_links, sender=linklist_cls.model)

        # 2. register listeners for the objects that are targets of Links,
        # only when get_absolute_url() is defined for the model
        if getattr(linklist_cls.model, 'get_absolute_url', None):
            model_signals.pre_save.disconnect(instance_pre_save, sender=linklist_cls.model)
            model_signals.post_save.disconnect(instance_post_save, sender=linklist_cls.model)
            model_signals.pre_delete.disconnect(do_instance_pre_delete, sender=linklist_cls.model)

    filebrowser.unregister_listeners()


@contextmanager
def enable_listeners(*args, **kwargs):
    register_listeners()
    try:
        yield
    finally:
        unregister_listeners()


@contextmanager
def disable_listeners(*args, **kwargs):
    unregister_listeners()
    try:
        yield
    finally:
        register_listeners()



import logging
import os.path
import re
from datetime import timedelta
from http import HTTPStatus
from urllib.parse import unquote, urlparse

import requests
from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.test.client import Client
from django.test.utils import modify_settings
from django.utils.encoding import iri_to_uri
from django.utils.functional import cached_property
from django.utils.timezone import now
from django.utils.translation import gettext as _
from requests.exceptions import ConnectionError, ReadTimeout

try:
    from reversion.revisions import revision_context_manager

    USE_REVERSION = True
except ImportError:
    USE_REVERSION = False

from .linkcheck_settings import (
    EXTERNAL_RECHECK_INTERVAL,
    EXTERNAL_REGEX_STRING,
    LINKCHECK_CONNECTION_ATTEMPT_TIMEOUT,
    MAX_URL_LENGTH,
    MEDIA_PREFIX,
    PROXIES,
    SITE_DOMAINS,
    TOLERATE_BROKEN_ANCHOR,
    TRUST_PROXY_SSL,
)

logger = logging.getLogger(__name__)


EXTERNAL_REGEX = re.compile(EXTERNAL_REGEX_STRING)


def html_decode(s):
    """
    Returns the ASCII decoded version of the given HTML string. This does
    NOT remove normal HTML tags like <p>.
    """
    html_codes = (("'", "&#39;"), ('"', "&quot;"), (">", "&gt;"), ("<", "&lt;"), ("&", "&amp;"))
    for code in html_codes:
        s = s.replace(code[1], code[0])
    return s


STATUS_CODE_CHOICES = [(s.value, f"{s.value} {s.phrase}") for s in HTTPStatus]
DEFAULT_USER_AGENT = f"{settings.SITE_DOMAIN} Linkchecker"
FALLBACK_USER_AGENT = (
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/90.0.4430.93 Safari/537.36"
)


class Url(models.Model):
    """
    Represents a distinct URL found somewhere in the models registered with linkcheck
    A single Url can have multiple Links associated with it.
    """

    # See http://www.boutell.com/newfaq/misc/urllength.html
    url = models.CharField(max_length=MAX_URL_LENGTH, unique=True)
    last_checked = models.DateTimeField(blank=True, null=True)
    anchor_status = models.BooleanField(null=True)
    ssl_status = models.BooleanField(null=True)
    status = models.BooleanField(null=True)
    status_code = models.IntegerField(choices=STATUS_CODE_CHOICES, null=True)
    redirect_status_code = models.IntegerField(choices=STATUS_CODE_CHOICES, null=True)
    message = models.CharField(max_length=1024, blank=True, null=True)
    error_message = models.CharField(max_length=1024, default="", blank=True)
    redirect_to = models.TextField(blank=True)

    @property
    def redirect_ok(self):
        return self.redirect_status_code < 300 if self.redirect_status_code else None

    @property
    def type(self):
        if self.external:
            return "external"
        if self.url.startswith("mailto:"):
            return "mailto"
        if self.url.startswith("tel:"):
            return "phone"
        elif self.internal_url == "":
            return "empty"
        elif self.internal_url.startswith("#"):
            return "anchor"
        elif self.internal_url.startswith(MEDIA_PREFIX):
            return "file"
        elif self.internal_url.startswith("/"):
            return "internal"
        else:
            return "invalid"

    @property
    def has_anchor(self):
        return "#" in self.url

    @property
    def anchor(self):
        return self.url.split("#")[1] if self.has_anchor else None

    @property
    def anchor_message(self):
        if not self.has_anchor or not self.last_checked:
            return ""
        if self.anchor == "":
            return _("Working empty anchor")
        if self.anchor_status is None:
            return _("Anchor could not be checked")
        elif self.anchor_status is False:
            return _("Broken anchor")
        return _("Working anchor")

    @property
    def ssl_message(self):
        if self.internal:
            return ""
        if self.external_url.startswith("http://"):
            return _("Insecure link")
        if self.ssl_status is None:
            return _("SSL certificate could not be checked")
        elif self.ssl_status is False:
            return _("Broken SSL certificate")
        return _("Valid SSL certificate")

    @property
    def get_message(self):
        if not self.last_checked and self.status is None:
            return _("URL Not Yet Checked")
        elif self.type == "empty":
            return _("Empty link")
        elif self.type == "invalid":
            return _("Invalid URL")
        elif self.type == "mailto":
            return "{} ({})".format(_("Email link"), _("not automatically checked"))
        elif self.type == "phone":
            return "{} ({})".format(_("Phone number link"), _("not automatically checked"))
        elif self.type == "anchor":
            return "{} ({})".format(_("Anchor link"), _("not automatically checked"))
        elif self.type == "file":
            return _("Working file link") if self.status else _("Missing file")
        elif not self.status_code:
            return self.error_message
        elif self.status_code < 300:
            return _("Working external link") if self.external else _("Working internal link")
        elif self.status_code < 400:
            permanent = self.status_code in [HTTPStatus.MOVED_PERMANENTLY, HTTPStatus.PERMANENT_REDIRECT]
            if self.redirect_ok:
                return _("Working permanent redirect") if permanent else _("Working temporary redirect")
            else:
                return _("Broken permanent redirect") if permanent else _("Broken temporary redirect")
        return _("Broken external link") if self.external else _("Broken internal link")

    @property
    def colour(self):
        if not self.last_checked:
            return "blue"
        elif self.status is True:
            return "green"
        else:
            return "red"

    def __str__(self):
        return self.url

    def __repr__(self):
        return f"<Url (id: {self.id}, url: {self.url})>"

    @cached_property
    def internal_url(self):
        """
        Remove current domain from URLs as the test client chokes when trying to test them during a page save
        They shouldn't generally exist but occasionally slip through
        If settings.SITE_DOMAINS isn't set then use settings.SITE_DOMAIN
        but also check for variants: example.org, www.example.org, test.example.org

        In case the URLs is external, `None` is returned.
        """

        # If the URL is not external, directly return it without processing
        if not EXTERNAL_REGEX.match(self.url):
            return self.url

        # May receive transformation before being checked
        prepared_url = self.url

        internal_exceptions = []
        if SITE_DOMAINS:  # If the setting is present
            internal_exceptions = SITE_DOMAINS
        elif getattr(settings, "SITE_DOMAIN", None):  # try using SITE_DOMAIN
            root_domain = settings.SITE_DOMAIN
            if root_domain.startswith("www."):
                root_domain = root_domain[4:]
            elif root_domain.startswith("test."):
                root_domain = root_domain[5:]
            internal_exceptions = [
                f"{protocol}://{sub}{root_domain}" for sub in ["", "www.", "test."] for protocol in ["http", "https"]
            ]

        for ex in internal_exceptions:
            if ex and prepared_url.startswith(ex):
                prepared_url = prepared_url.replace(ex, "", 1)

        # If the URL is still external, return `None`
        if EXTERNAL_REGEX.match(prepared_url):
            return None

        logger.debug("Internal URL: %s", prepared_url)
        return prepared_url

    @cached_property
    def external_url(self):
        """
        Prepare an external URL to be checked with requests:
        - Remove hash anchors
        - Ensure correct encoding
        """
        # If the URL is internal, return `None`
        if self.internal:
            return None

        # Encode path and query and remove anchor fragment
        parsed = urlparse(self.url)
        external_url = parsed._replace(
            path=iri_to_uri(parsed.path), query=iri_to_uri(parsed.query), fragment=""
        ).geturl()

        logger.debug("External URL: %s", external_url)
        return external_url

    @property
    def internal(self):
        """
        Check whether this URL is internal
        """
        return self.internal_url is not None

    @property
    def external(self):
        """
        Check whether this URL is external
        """
        return not self.internal

    def reset_for_check(self):
        """
        Reset all fields which depend on the status after checking a URL.
        This is done to ensure that results from the last check do not remain if the fields are not overwritten.
        """
        # Reset all database fields
        self.anchor_status = None
        self.status = None
        self.status_code = None
        self.redirect_status_code = None
        self.ssl_status = None
        self.error_message = ""
        self.message = ""

    def check_url(self, check_internal=True, check_external=True, external_recheck_interval=EXTERNAL_RECHECK_INTERVAL):
        """
        Return:
         * True if the link was checked and found valid
         * False if the link was checked and found invalid
         * None if the link was not checked
        """

        if check_internal and self.internal:
            return self.check_internal()
        elif check_external and self.external:
            return self.check_external(external_recheck_interval)
        else:
            return None

    def check_internal(self):
        """
        Check an internal URL
        """
        if not self.internal:
            logger.info("URL %r is not internal", self)
            return None

        logger.debug("checking internal link: %s", self.internal_url)

        # Reset all fields in case they were already set
        self.reset_for_check()

        from linkcheck.utils import LinkCheckHandler

        if self.type == "empty":
            self.status = False
            self.message = "Empty link"

        elif self.type == "mailto":
            self.message = "Email link (not automatically checked)"

        elif self.type == "phone":
            self.message = "Phone number (not automatically checked)"

        elif self.type == "anchor":
            self.message = "Link to within the same page (not automatically checked)"

        elif self.type == "file":
            # TODO: Assumes a direct mapping from media url to local filesystem path.
            # This will break quite easily for alternate setups
            path = settings.MEDIA_ROOT + unquote(self.internal_url)[len(MEDIA_PREFIX) - 1:]
            decoded_path = html_decode(path)
            self.status = os.path.exists(path) or os.path.exists(decoded_path)
            self.message = "Working file link" if self.status else "Missing Document"

        elif self.type == "internal":
            old_prepend_setting = settings.PREPEND_WWW
            settings.PREPEND_WWW = False
            c = Client()
            c.handler = LinkCheckHandler()
            with modify_settings(ALLOWED_HOSTS={"append": "testserver"}):
                response = c.get(self.internal_url)
            self.status_code = response.status_code
            if response.status_code < 300:
                self.message = "Working internal link"
                self.status = True
            elif response.status_code < 400:
                initial_location = response.get("Location")
                redirect_type = "permanent" if response.status_code == 301 else "temporary"
                with modify_settings(ALLOWED_HOSTS={"append": "testserver"}):
                    response = c.get(self.internal_url, follow=True)
                if response.redirect_chain:
                    self.redirect_to, _ = response.redirect_chain[-1]
                else:
                    self.redirect_to = initial_location
                self.redirect_status_code = response.status_code
                self.status = response.status_code < 300
                redirect_result = "Working" if self.status else "Broken"
                self.message = f"{redirect_result} {redirect_type} redirect"
            else:
                self.status = False
                self.message = "Broken internal link"

            # Check the anchor (if it exists)
            self.check_anchor(response.content)

            settings.PREPEND_WWW = old_prepend_setting
        else:
            self.status = False
            self.message = "Invalid URL"

        if USE_REVERSION:
            # using test client will clear the RevisionContextManager stack.
            revision_context_manager.start()

        self.last_checked = now()
        self.save()
        return self.status

    def check_external(self, external_recheck_interval=EXTERNAL_RECHECK_INTERVAL):
        """
        Check an external URL
        """
        if not self.external:
            logger.info("URL %r is not external", self)
            return None

        logger.info("checking external link: %s", self.url)
        external_recheck_datetime = now() - timedelta(minutes=external_recheck_interval)

        if self.last_checked and (self.last_checked > external_recheck_datetime):
            logger.debug(
                "URL was last checked in the last %s minutes, so not checking it again", external_recheck_interval
            )
            return self.status

        # Reset all fields in case they were already set
        self.reset_for_check()

        request_params = {
            "allow_redirects": True,
            "headers": {"User-Agent": DEFAULT_USER_AGENT},
            "timeout": LINKCHECK_CONNECTION_ATTEMPT_TIMEOUT,
            "verify": True,
        }
        if PROXIES:
            request_params["verify"] = not TRUST_PROXY_SSL
            request_params["proxies"] = PROXIES

        try:
            try:
                # At first try a HEAD request
                fetch = requests.head
                response = fetch(self.external_url, **request_params)
                # If no exceptions occur, the SSL certificate is valid
                if self.external_url.startswith("https://"):
                    self.ssl_status = True
            except ConnectionError as e:
                # This error could also be caused by an incomplete root certificate bundle,
                # so let's retry without verifying the certificate
                if "unable to get local issuer certificate" in str(e):
                    request_params["verify"] = False
                    response = fetch(self.external_url, **request_params)
                else:
                    # Re-raise exception if it's definitely not a false positive
                    raise
            # If HEAD is not allowed, let's try with GET
            if response.status_code in [HTTPStatus.BAD_REQUEST, HTTPStatus.METHOD_NOT_ALLOWED]:
                logger.debug("HEAD is not allowed, retry with GET")
                fetch = requests.get
                response = fetch(self.external_url, **request_params)
            # If access is denied, possibly the user agent is blocked
            if response.status_code == HTTPStatus.FORBIDDEN:
                logger.debug("Forbidden, retry with different user agent")
                request_params["headers"] = {"User-Agent": FALLBACK_USER_AGENT}
                response = fetch(self.external_url, **request_params)
            # If URL contains hash anchor and is a valid HTML document, let's repeat with GET
            elif (
                self.has_anchor
                and response.ok
                and fetch == requests.head
                and "text/html" in response.headers.get("content-type")
            ):
                logger.debug("Retrieve content for anchor check")
                fetch = requests.get
                response = fetch(self.external_url, **request_params)
        except ReadTimeout:
            self.status = False
            self.message = "Other Error: The read operation timed out"
            self.error_message = "The read operation timed out"
        except ConnectionError as e:
            self.status = False
            self.message = self.error_message = format_connection_error(e)
            if "SSLError" in str(e):
                self.ssl_status = False
        except Exception as e:
            self.status = False
            self.message = f"Other Error: {e}"
            self.error_message = str(e)
        else:
            self.status = response.status_code < 300
            self.message = f"{response.status_code} {response.reason}"
            logger.debug("Response message: %s", self.message)

            # If initial response was a redirect, return the initial return code
            if response.history:
                logger.debug("Redirect history: %r", response.history)
                if response.ok:
                    self.message = f"{response.history[0].status_code} {response.history[0].reason}"
                self.redirect_to = response.url
                self.redirect_status_code = response.status_code
                self.status_code = response.history[0].status_code
            else:
                self.status_code = response.status_code

            # Check the anchor (if it exists)
            if fetch == requests.get:
                self.check_anchor(response.text)
            if not request_params["verify"]:
                self.message += ", SSL certificate could not be verified"

        # When a rate limit was hit or the server returned an internal error, do not update
        # the last_checked date so the result is not cached for EXTERNAL_RECHECK_INTERVAL minutes
        if not self.status_code or self.status_code != HTTPStatus.TOO_MANY_REQUESTS and self.status_code < 500:
            self.last_checked = now()
        self.save()
        return self.status

    def check_anchor(self, html):
        from linkcheck import parse_anchors

        scope = "internal" if self.internal else "external"

        # Only check when the URL contains an anchor
        if self.has_anchor:
            # Empty fragment '#' is always valid
            if not self.anchor:
                self.anchor_status = True
                self.message += f", working {scope} hash anchor"
            else:
                try:
                    names = parse_anchors(html)
                # Known possible errors include: AssertionError, NotImplementedError, UnicodeDecodeError
                except Exception as e:
                    logger.debug("%s while parsing anchors: %s", type(e).__name__, e)
                    self.message += ", failed to parse HTML for anchor"
                    if not TOLERATE_BROKEN_ANCHOR:
                        self.status = False
                else:
                    if self.anchor in names:
                        self.anchor_status = True
                        self.message += f", working {scope} hash anchor"
                    else:
                        self.anchor_status = False
                        self.message += f", broken {scope} hash anchor"
                        if not TOLERATE_BROKEN_ANCHOR:
                            self.status = False
        return self.anchor_status, self.anchor_message


class Link(models.Model):
    """
    A Link represents a specific URL in a specific field in a specific model
    It can be come from a single field such as a URLField or a field containing multiple links
    Such as a HTML or Rich Text field.
    Multiple Links can reference a single Url
    """

    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey("content_type", "object_id")
    field = models.CharField(max_length=128)
    url = models.ForeignKey(Url, related_name="links", on_delete=models.CASCADE)
    text = models.CharField(max_length=256, default="")
    ignore = models.BooleanField(default=False)

    class Meta:
        indexes = [
            models.Index(fields=["content_type", "object_id"], name="content_type_and_object_id"),
        ]

    @property
    def display_url(self):
        # when page /test/ has a anchor link to /test/#anchor, we display it
        # as "#anchor" rather than "/test/#anchor"
        if self.url.url.count("#") and hasattr(self.content_object, "get_absolute_url"):
            url_part, anchor_part = self.url.url.split("#")
            absolute_url = self.content_object.get_absolute_url()
            if url_part == absolute_url:
                return "#" + anchor_part
        return self.url.url

    def __str__(self):
        return f"{self.url.url} ({self.content_object})"

    def __repr__(self):
        return f"<Link (id: {self.id}, url: {self.url!r}, source: {self.content_object!r})>"


def link_post_delete(sender, instance, **kwargs):
    try:
        # url.delete() => link.delete() => link_post_delete
        # in this case link.url is already deleted from db, so we need a try here.
        url = instance.url
        count = url.links.all().count()
        if count == 0:
            logger.debug("This was the last link for %r, so deleting it", url)
            url.delete()
    except Url.DoesNotExist:
        pass


def format_connection_error(e):
    """
    Helper function to provide better readable output of connection errors
    """
    # If the exception message is wrapped in an "HTTPSConnectionPool", only give the underlying cause
    reason = re.search(r"\(Caused by ([a-zA-Z]+\(.+\))\)", str(e))
    if not reason:
        return f"Connection Error: {e}"
    reason = reason[1]
    # If the underlying cause is a new connection error, provide additional formatting
    if reason.startswith("NewConnectionError"):
        return format_new_connection_error(reason)
    # If the underlying cause is a name resolution error, provide additional formatting
    if reason.startswith("NameResolutionError"):
        return format_name_resolution_error(reason)
    # If the underlying cause is an SSL error, provide additional formatting
    if reason.startswith("SSLError"):
        return format_ssl_error(reason)
    return f"Connection Error: {reason}"


def format_new_connection_error(reason):
    """
    Helper function to provide better readable output of new connection errors thrown by urllib3
    """
    connection_reason = re.search(
        r"NewConnectionError\('<urllib3\.connection\.HTTPSConnection object at 0x[0-9a-f]+>: (.+)'\)",
        reason,
    )
    if connection_reason:
        return f"New Connection Error: {connection_reason[1]}"
    return reason


def format_name_resolution_error(reason):
    """
    Helper function to provide better readable output of name resolution errors thrown by urllib3
    """
    resolution_reason = re.search(
        r"NameResolutionError\([\"']<urllib3\.connection\.HTTPSConnection object at 0x[0-9a-f]+>: (.+)[\"']\)",
        reason,
    )
    if resolution_reason:
        return f"Name Resolution Error: {resolution_reason[1]}"
    return reason


def format_ssl_error(reason):
    """
    Helper function to provide better readable output of SSL errors thrown by urllib3
    """
    ssl_reason = re.search(r"SSLError\([a-zA-Z]+\((.+)\)\)", reason)
    if ssl_reason:
        # If the reason lies withing the ssl c library, hide additional debug output
        ssl_c_reason = re.search(r"1, '\[SSL: [A-Z\d_]+\] (.+) \(_ssl\.c:\d+\)'", ssl_reason[1])
        if ssl_c_reason:
            return f"SSL Error: {ssl_c_reason[1]}"
        return f"SSL Error: {ssl_reason[1]}"
    return reason



from celery import shared_task
from django.apps import apps

from .worker_tasks import do_check_instance_links, do_instance_post_save


@shared_task
def do_check_link(app: str, model: str, id: int):
    sender = apps.get_app_config(app).get_model(model)

    linklist_cls = sender._linklist

    instance = sender.objects.get(id=id)

    do_check_instance_links(sender, instance, linklist_cls)


@shared_task
def do_post_save(app: str, model: str, id: int, **kwargs):
    sender = apps.get_app_config(app).get_model(model)

    linklist_cls = sender._linklist

    instance = sender.objects.get(id=id)

    do_instance_post_save(sender, instance, linklist_cls, **kwargs)



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('linkcheck', '0009_url_add_ssl_status'),
    ]

    operations = [
        migrations.AddField(
            model_name='url',
            name='error_message',
            field=models.CharField(blank=True, default='', max_length=1024),
        ),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("linkcheck", "0007_url_add_redirect_status_code"),
    ]

    operations = [
        migrations.AddField(
            model_name="url",
            name="anchor_status",
            field=models.BooleanField(null=True),
        ),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("linkcheck", "0008_url_add_anchor_status"),
    ]

    operations = [
        migrations.AddField(
            model_name="url",
            name="ssl_status",
            field=models.BooleanField(null=True),
        ),
    ]



from itertools import groupby
from operator import itemgetter

from django import forms
from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.core.paginator import Paginator
from django.http import JsonResponse
from django.shortcuts import render
from django.templatetags.static import static
from django.urls import NoReverseMatch, reverse
from django.utils.translation import gettext as _

from linkcheck import update_lock
from linkcheck.linkcheck_settings import RESULTS_PER_PAGE
from linkcheck.models import Link
from linkcheck.utils import get_coverage_data


@staff_member_required
def coverage(request):

    coverage_data = get_coverage_data()

    if request.GET.get('config', False):
        # Just render the suggested linklist code
        template = 'linkcheck/suggested_configs.html'
        context = {'coverage_data': [x['suggested_config'] for x in coverage_data]}
    else:
        # Render a nice report
        template = 'linkcheck/coverage.html'
        context = {'coverage_data': coverage_data}

    return render(request, template, context)


@staff_member_required
def report(request):

    outerkeyfunc = itemgetter('content_type_id')
    content_types_list = []

    if request.method == 'POST':

        ignore_link_id = request.GET.get('ignore', None)
        if ignore_link_id is not None:
            link = Link.objects.get(id=ignore_link_id)
            link.ignore = True
            link.save()
            if is_ajax(request):
                json_data = {'link': link.pk}
                return JsonResponse(json_data)

        unignore_link_id = request.GET.get('unignore', None)
        if unignore_link_id is not None:
            link = Link.objects.get(id=unignore_link_id)
            link.ignore = False
            link.save()
            if is_ajax(request):
                json_data = {'link': link.pk}
                return JsonResponse(json_data)

        recheck_link_id = request.GET.get('recheck', None)
        if recheck_link_id is not None:
            link = Link.objects.get(id=recheck_link_id)
            url = link.url
            url.check_url(external_recheck_interval=0)
            links = [x[0] for x in url.links.values_list('id')]
            if is_ajax(request):
                json_data = ({
                    'links': links,
                    'message': url.message,
                    'colour': url.colour,
                })
                return JsonResponse(json_data)

    link_filter = request.GET.get('filters', 'show_invalid')

    qset = Link.objects.order_by('-url__last_checked')
    if link_filter == 'show_valid':
        qset = qset.filter(ignore=False, url__status__exact=True)
        report_type = _('Valid links')
    elif link_filter == 'show_unchecked':
        qset = qset.filter(ignore=False, url__last_checked__exact=None)
        report_type = _('Untested links')
    elif link_filter == 'ignored':
        qset = qset.filter(ignore=True)
        report_type = _('Ignored links')
    else:
        qset = qset.filter(ignore=False, url__status__exact=False)
        report_type = _('Broken links')

    paginated_links = Paginator(qset, RESULTS_PER_PAGE, 0, True)

    try:
        page = int(request.GET.get("page", "1"))
    except ValueError:
        page = 0
    # offset = (page - 1) * RESULTS_PER_PAGE
    links = paginated_links.page(page)

    # This code groups links into nested lists by content type and object id
    # It's a bit nasty but we can't use groupby unless be get values()
    # instead of a queryset because of the 'Object is not subscriptable' error

    t = sorted(links.object_list.values(), key=outerkeyfunc)
    for tk, tg in groupby(t, outerkeyfunc):
        innerkeyfunc = itemgetter('object_id')
        objects = []
        tg = sorted(tg, key=innerkeyfunc)
        for ok, og in groupby(tg, innerkeyfunc):
            content_type = ContentType.objects.get(pk=tk)
            og = list(og)
            try:
                object = None
                if content_type.model_class():
                    object = content_type.model_class().objects.get(pk=ok)
            except ObjectDoesNotExist:
                pass
            try:
                admin_url = object.get_admin_url()  # TODO allow method name to be configurable
            except AttributeError:
                try:
                    admin_url = reverse(f'admin:{content_type.app_label}_{content_type.model}_change', args=[ok])
                except NoReverseMatch:
                    admin_url = None

            objects.append({
                'object': object,
                # Convert values_list back to queryset. Do we need to get values() or do we just need a list of ids?
                'link_list': Link.objects.in_bulk([x['id'] for x in og]).values(),
                'admin_url': admin_url,
            })
        content_types_list.append({
            'content_type': content_type,
            'object_list': objects
        })

    # Pass any querystring data back to the form minus page
    rqst = request.GET.copy()
    if 'page' in rqst:
        del rqst['page']

    return render(request, 'linkcheck/report.html', {
            'content_types_list': content_types_list,
            'pages': links,
            'filter': link_filter,
            'media': forms.Media(js=[static(get_jquery_min_js())]),
            'qry_data': rqst.urlencode(),
            'report_type': report_type,
            'ignored_count': Link.objects.filter(ignore=True).count(),
        },
    )


def get_jquery_min_js():
    """
    Return the location of jquery.min.js. It's an entry point to adapt the path
    when it changes in Django.
    """
    return 'admin/js/vendor/jquery/jquery.min.js'


def get_status_message():
    if update_lock.locked():
        return "Still checking. Please refresh this page in a short while. "
    else:
        broken_links = Link.objects.filter(ignore=False, url__status=False).count()
        if broken_links:
            return (
                "<span style='color: red;'>We've found {} broken link{}.</span><br>"
                "<a href='{}'>View/fix broken links</a>".format(
                    broken_links,
                    "s" if broken_links > 1 else "",
                    reverse('linkcheck_report'),
                )
            )
        else:
            return ''


def is_ajax(request):
    return request.META.get('HTTP_X_REQUESTED_WITH') == 'XMLHttpRequest'



from django.db import migrations, models

from linkcheck.models import STATUS_CODE_CHOICES


class Migration(migrations.Migration):

    dependencies = [
        ("linkcheck", "0005_default_big_auto_field"),
    ]

    operations = [
        migrations.AddField(
            model_name="url",
            name="status_code",
            field=models.IntegerField(
                choices=STATUS_CODE_CHOICES,
                null=True,
            ),
        ),
    ]



import os
import subprocess

from setuptools import build_meta as default
from setuptools.build_meta import *  # noqa: F401, F403


def compile_translation_files():
    print("Compiling translation files...")
    subprocess.run(["django-admin", "compilemessages"], cwd="linkcheck")


def should_compile_translation_files():
    skip_translations = os.environ.get("LINKCHECK_SKIP_TRANSLATIONS")
    if skip_translations and skip_translations.lower() in ("1", "true", "yes", "t", "y"):
        return False

    return True


def build_sdist(sdist_directory, config_settings=None):
    if should_compile_translation_files():
        compile_translation_files()

    return default.build_sdist(sdist_directory, config_settings)


def build_wheel(wheel_directory, config_settings=None, metadata_directory=None):
    if should_compile_translation_files():
        compile_translation_files()

    return default.build_wheel(
        wheel_directory,
        config_settings=config_settings,
        metadata_directory=metadata_directory,
    )



# Generated by Django 5.1.3 on 2024-11-25 18:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('linkcheck', '0010_url_add_error_message'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='link',
            index=models.Index(fields=['content_type', 'object_id'], name='content_type_and_object_id'),
        ),
    ]



# This file works with our fork of django-cron.
# It's use is optional
# Use any means you like to run scheduled jobs.
#
# Note - you only need to run scheduled jobs if you want to check external links
# that may have died since the link was last edited
#
# Links are checked via signals any time a link-containing object is saved by Django

from django_cron import WEEK, Job, cronScheduler

from linkcheck.linkcheck_settings import (
    EXTERNAL_RECHECK_INTERVAL,
    MAX_CHECKS_PER_RUN,
)
from linkcheck.utils import check_links, find_all_links


class RunLinkCheckFind(Job):

    run_every = WEEK

    def job(self):
        find_all_links()


cronScheduler.register(RunLinkCheckFind)


class RunLinkCheckInternal(Job):

    run_every = WEEK

    def job(self):
        check_links(limit=MAX_CHECKS_PER_RUN, check_external=False)


cronScheduler.register(RunLinkCheckInternal)


class RunLinkCheckExternal(Job):

    run_every = WEEK

    def job(self):
        check_links(
            external_recheck_interval=EXTERNAL_RECHECK_INTERVAL,
            limit=MAX_CHECKS_PER_RUN,
            check_internal=False,
        )


cronScheduler.register(RunLinkCheckExternal)



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('linkcheck', '0002_url_redirect_to'),
    ]

    operations = [
        migrations.AlterField(
            model_name='url',
            name='redirect_to',
            field=models.TextField(blank=True),
        ),
    ]



from django import http
from django.contrib import admin
from django.urls import include, path
from django.views.generic import RedirectView

from linkcheck.tests.sampleapp import views


def handler404(*args, **kwargs):
    return http.HttpResponseNotFound("")


urlpatterns = [
    path('admin/linkcheck/', include('linkcheck.urls')),
    path('admin/', admin.site.urls),
    path('public/', views.http_response, {'code': '200'}),
    path('http/<int:code>/', views.http_response),
    path('http/<int:code>/rückmeldung/', views.http_response),
    path('http/getonly/<int:code>/', views.http_response_get_only),
    path('http/block-user-agent/', views.http_block_user_agent),
    path('http/block-user-agent/block-head/', views.http_block_user_agent, {'block_head': True}),
    path('http/redirect/<int:code>/', views.http_redirect),
    path('http/redirect_to_404/', views.http_redirect_to_404),
    path('http/redirect_to_anchor/', views.http_redirect_to_anchor),
    path('http/brokenredirect/', RedirectView.as_view(url='/non-existent/')),
    path('http/anchor/', views.http_response_with_anchor),
    path('timeout/', views.timeout),
    path('static-files/video.mp4', views.static_video),
    path('static-files/fake-video.mp4', views.static_video_forged_content_type),
]
